#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include "ddos.h"

#define OP_LIST 1
#define OP_EXTRACT 2

int verbose = 0;
int operation = OP_LIST;

char *createfn(struct dentry *tocrec) {
    static char buffer[16];
    memset(buffer, 0, sizeof buffer);
    strncpy(buffer, tocrec->name, sizeof tocrec->name);
    strcat(buffer, ".");
    strncat(buffer, tocrec->ext, sizeof tocrec->ext);
    return buffer;
}

int getsize(struct allocblock *b, int num, int last) {
    int sum = 0;
    int i;

    for (i = 0; i < num; i++) {
        sum += b[i].count * 256;
    }
    sum = sum - 256 + last;
    return sum;
}

/*
 * Return an entry in the directory or NULL if
 * end of dir. Does not return deleted entries.
 */
struct dentry *getentry(FILE *fp, int *progress) {

    static struct dentry tocrec;
    int eod, deleted, contentry, conted;
    char unused[6];

    do {
        fread(&tocrec, sizeof tocrec, 1, fp);
        eod = (tocrec.flag & 8) != 0;
        deleted = (tocrec.flag & 128) != 0;
        contentry = tocrec.flag & 1;
        if (eod) {
            return NULL;
        }
        (*progress)++;
        if (*progress == 10) {
            fread(&unused, sizeof unused, 1, fp); // Skip unused bytes
            *progress = 0;
        }
        conted = tocrec.flag & 32;
        if (!contentry && !deleted)
            return &tocrec;
    } while (contentry || deleted);
}

/*
 * List the directory of a DragonDos formatted disk.
 * FIXME: Does not calculate size of files with continuations correctly.
 */
void listdir(FILE *fp, struct diskinfo *h) {
    int i, sector, recnum;
    struct dentry *tocrec;
    //char namebuffer[16];
    char *namebuffer;

    seektodir(fp, h);
    recnum = 0;
    for (sector = 0; sector < 18; sector++) {
        for (i = 0; i < 10; i++) {
            tocrec = getentry(fp, &recnum);
            if (tocrec == NULL) {
                return;
            }
            namebuffer = createfn(tocrec);
            if (verbose)
                printf("%-14s%6d\n", namebuffer, getsize(tocrec->alloc, 4, tocrec->last));
            else
                printf("%-14s\n", namebuffer);

        }
    }
}

/*
 * Open the disk and list the directory.
 */
void listtoc(char *filename) {
    FILE *fp;
    struct diskinfo h;

    if (verbose) {
        printf("%s\n", filename);
        printf("Filename        Size\n");
        printf("====================\n");
    }
    fp = fopen(filename, "r");
    vdkhead(fp, &h);
    listdir(fp, &h);
    printf("\n");
    fclose(fp);
}

void usage(char *prgname) {
    fprintf(stderr, "Usage: [-v] %s <diskimage>\n", prgname);
    exit(1);
}

int main(int argc, char *argv[]) {

    int opt;
    int i, sz;

    if (argc < 2) {
        fprintf(stderr, "Usage: %s disk_image\n", argv[0]);
        exit(1);
    }
    while ((opt = getopt(argc, argv, "v")) != -1) {
        switch(opt) {
        case 'v':
            verbose = 1;
            break;
        default:
            usage(argv[0]);
        }
    }

    if (optind >= argc) {
        fprintf(stderr, "Expected argument after options\n");
        exit(EXIT_FAILURE);
    }

    switch (operation) {
    case OP_LIST:
    default:
        listtoc(argv[optind]);
        break;
    }
    exit(0);
}

