/*
 * Extract a file from a Dragon Dos disk
 * Bugs:
 *   Does not handle programs with two continuation segments.
 *   Should not use goto.
 */
#include "ddos.h"
#include <string.h>

static char *curroutfn;
static FILE *fp;
static struct diskinfo *h;

char *createfn(struct dentry *tocrec) {
    static char buffer[12];
    memset(buffer, 0, sizeof buffer);
    strncpy(buffer, tocrec->name, 8);
    strcat(buffer, ".");
    strncat(buffer, tocrec->ext, 3);
    return buffer;
}

#define SECTORSIZE 256
#define CONTENTRY 1
#define PROTECTED 2
#define EODATA    8
#define CONTINUED 32
#define DELETED   128

void copysectors(FILE *outfp, int sectors, int howmany) {
    int c,j,k,siz;

    for (j = 1; j <= sectors; j++) {
        siz = SECTORSIZE;
        if (j == sectors) siz = howmany;

        for (k = 0; k < siz; k++) {
            c = getc(fp);
            putc(c, outfp);
        }
    }
}


void writefile(struct allocblock *segments, int finish) {
    FILE *outfp;
    int i, howmany, sectors;

    outfp = fopen(curroutfn, "w");
    for (i = 0; i < 11; i++) {
        if (segments[i].count == 0)
            break;
        long segloc = 0L + (segments[i].high * 256 + segments[i].low) * SECTORSIZE
                            + h->vdksize;
        fseek(fp, segloc, SEEK_SET);
        if (segments[i+1].high == 0 && segments[i+1].low == 0 && finish != 0) {
            howmany = finish;
        } else {
            howmany = SECTORSIZE;
        }
        sectors = segments[i].count;
        copysectors(outfp, sectors, howmany);
    }
    fclose(outfp);
}


void findintoc(char *filename) {
    int recnum, finish;
    struct dentry tocrec;
    struct centry crec;
    struct allocblock segments[12];
    char unused[6];
    int eod, deleted, contentry;
    int i;
    int contseg;
    char *dosname;

    seektodir(fp, h);
    recnum = 0;
    do {
        fread(&tocrec, sizeof tocrec, 1, fp);
        eod = (tocrec.flag & EODATA) != 0;
        if (eod) break;
        deleted = (tocrec.flag & DELETED) != 0;
        if (deleted) {
            goto nextentry;
        }
        contentry = tocrec.flag & CONTENTRY;
        if (eod) {
            break;
        }
        if (contentry && contseg == recnum) {
            memcpy(&crec, &tocrec, sizeof tocrec); /* Convert to centry format */
            for (i = 0; i < 7; i++) {
                memcpy(&segments[i + 4], &crec.alloc[i], sizeof(struct allocblock));
                contseg = crec.last;
            }
        }
        if (contentry) {
            goto nextentry;
        }
        if (!contentry && !deleted) {
            dosname = createfn(&tocrec);
            if (strcasecmp(filename, dosname) == 0) {
                memset(&segments, 0, sizeof segments);
                for (i = 0; i < 4; i++) {
                    memcpy(&segments[i], &tocrec.alloc[i], sizeof(struct allocblock));
                }
                finish = tocrec.last & 0xFF;  /* Number of bytes in last sector */
                if (tocrec.flag & CONTINUED != 0) {
                    contseg = tocrec.last;
                }
            }
        }
nextentry: recnum += 1;
        if (recnum % 10 == 0) {
            fread(&unused, sizeof unused, 1, fp);
        }
    } while (!eod);
    if (segments[0].count == 0) {
        fprintf(stderr, "File not found\n");
    } else {
        writefile(segments, finish);
    }
}


int main(int argc, char *argv[]) {
    int i, sz;
    struct diskinfo hrec;

    h = &hrec;
    if (argc != 3) {
        fprintf(stderr, "Usage: %s <disk> <file>\n", argv[0]);
        exit(1);
    }
    curroutfn = argv[2];
    fp = fopen(argv[1], "r");
    vdkhead(fp, h);
    findintoc(argv[2]);
    fclose(fp);
}

